from utils.embeddings import one_hot_embedding
import torch


def performance_multi(logits, real):
    logits = logits.max(1)[1]
    real = real.contiguous().view(-1)
    non_pad_mask = real.ne(0)
    n_correct = logits.eq(real)
    n_correct = n_correct.masked_select(non_pad_mask).sum().item()

    return n_correct


def performance_regression(logits, real, threshold):
    ''' Apply label smoothing if needed '''

    logits = logits.ge(threshold)
    real = real.contiguous().view(-1)
    non_pad_mask = real.ne(0)
    n_correct = logits.eq(real)
    n_correct = n_correct.masked_select(non_pad_mask).sum().item()

    return n_correct


def accuracy(pred, real, threshold=None):
    if threshold==None:
        n = real.shape[0]
        pred = pred.argmax(dim=-1).view(n, -1)
        real = real.view(n, -1)
        acc = pred.byte() == real.byte()
        acc = (acc).float().mean()
        return acc
    else:
        acc = (pred > threshold).byte() == real.byte()
        acc = (acc).float().mean()
        return acc



def confusion_matrix(pred, real, threshold=None):
    '''
        For binary or multi class classification
        If pred is probabilities, convert to binary form
    '''
    dim = real.shape[-1]
    if dim == 1:
        if threshold != None:
            pred = (pred > threshold).byte()
            pred = one_hot_embedding(pred, 2)
            real = one_hot_embedding(real, 2)
        else:
            pred = one_hot_embedding(pred, real.max(-1))
            real = one_hot_embedding(real, real.max(-1))

    cm = torch.bmm(real.t(), pred)
    return cm


def precision_recall(pred, real, d_output, threshold=None, average='macro', eps = 1e-6):
    n = real.shape[0]
    dim = real.shape[-1]
    if dim == 1:
        if threshold != None:
            pred = (pred > threshold).byte()
            pred = one_hot_embedding(pred, 2)
            real = one_hot_embedding(real, 2)
        else:
            pred = one_hot_embedding(torch.argmax(pred, -1).view(n, -1), d_output)
            real = one_hot_embedding(real, d_output)

    tp = (pred & real)
    tp_count = tp.sum(0).float()
    fp_count = (pred - tp).sum(0).float() + eps
    fn_count = (real - tp).sum(0).float() + eps
    precision = tp_count / (tp_count + fp_count)
    recall = tp_count / (tp_count + fn_count)

    if threshold != None:
        return precision, recall, precision, recall

    else:
        if average == 'macro':
            precision_avg = precision.mean()
            recall_avg = precision.mean()

        else:
            precision_avg = tp.sum(0).sum() / n
            recall_avg = precision_avg

    return precision, recall, precision_avg, recall_avg

class Evaluator():
    def __init__(self):
        self.total_loss = 0
        self.total_acc = 0
        self.total_pre = 0
        self.total_rec = 0
        self.batch_counter = 0


    def __call__(self, loss, acc, precision, recall):
        self.total_loss += loss
        self.total_acc += acc
        self.total_pre += precision
        self.total_rec += recall
        self.batch_counter += 1

    def avg_results(self):
        loss_avg = self.total_loss / self.batch_counter
        acc_avg = self.total_acc / self.batch_counter
        pre_avg = self.total_pre / self.batch_counter
        rec_avg = self.total_rec / self.batch_counter
        return loss_avg, acc_avg, pre_avg, rec_avg


import torch.nn as nn
from Layers.plain import PlainConvLayer
class PlainConvStack(nn.Module):
    ''' Self Attention Stack Module '''

    def __init__(
            self, expansion, d_features, n_layers, n_channels, dropout=0.1, use_bottleneck=None, d_bottleneck=None):
        super().__init__()
        self.layer_stack = nn.ModuleList([
            PlainConvLayer(n_channels, d_features, dropout)
            for _ in range(n_layers)])

    def forward(self, x):
        for layers in self.layer_stack:
            x = layers(
                x)

        return x


import logging
class logger():
    def __init__(self, logger_name, log_path, mode, level=logging.INFO, format = "%(asctime)s - %(message)s"):
        self.logging = logging.getLogger(logger_name)
        self.logging.setLevel(level)
        fh = logging.FileHandler(log_path, mode)
        fh.setLevel(level)
        sh = logging.StreamHandler()
        sh.setLevel(logging.INFO)
        formatter = logging.Formatter(format)
        fh.setFormatter(formatter)
        sh.setFormatter(formatter)
        self.logging.addHandler(fh)
        self.logging.addHandler(sh)

    def info(self, msg):
        return self.logging.info(msg)





import torch
import torch.nn as nn
import numpy as np

def get_sinusoid_encoding_table(n_position, d_hid, padding_idx=None):
    '''
        Arguments:
            n_position {Int} -- the maximum position
            d_hid {Int} -- the dimension of the embedding
            padding_idx -- padding symbol

        Returns:
            sinusoid_table {Tensor, shape: [n_position+1, d_hid]} -- sinusoid encoding table
    '''
    n_position += 1

    def cal_angle(position, hid_idx):
        return position / np.power(10000, 2 * (hid_idx // 2) / d_hid)

    def get_posi_angle_vec(position):
        return [cal_angle(position, hid_j) for hid_j in range(d_hid)]

    sinusoid_table = np.array([get_posi_angle_vec(pos_i) for pos_i in range(n_position)])

    sinusoid_table[:, 0::2] = np.sin(sinusoid_table[:, 0::2])  # dim 2i
    sinusoid_table[:, 1::2] = np.cos(sinusoid_table[:, 1::2])  # dim 2i+1

    if padding_idx is not None:
        # zero vector for padding dimension
        sinusoid_table[padding_idx] = 0.

    return torch.FloatTensor(sinusoid_table)

class SinusoidPositionEncoding(nn.Module):

    def __init__(self, d_features, max_length=None, d_meta = None):
        super().__init__()
        self.position_enc = nn.Embedding.from_pretrained(
            get_sinusoid_encoding_table(max_length, d_features, padding_idx=0),
            freeze=True)

    def forward(self, x):
        '''
            Argument:
                x {Tensor, shape: [batch, length]} -- sequence position index masked
            Returns:
                x {Tensor, shape: [batch, length, d_features]} -- positional encoding
        '''
        x = self.position_enc(x)
        pass
        return x

class LinearPositionEncoding(nn.Module):

    def __init__(self, d_features, max_length=None, d_meta = None):
        super().__init__()
        self.position_enc = nn.Linear(d_meta, d_features, bias=False)
        self.tanh = nn.Tanh()

    def forward(self, x):
        x = self.position_enc(x)
        x = self.tanh(x)
        return x


class TimeFacilityEncoding(nn.Module):

    def __init__(self, d_features, max_length=None, d_meta = None):
        super().__init__()
        self.time_enc = nn.Embedding.from_pretrained(
            get_sinusoid_encoding_table(max_length, d_features, padding_idx=0),
            freeze=True)
        torch.manual_seed(1)
        self.facility_enc = torch.nn.Embedding(d_meta+1, d_features, padding_idx=0)


    def forward(self, x):
        '''
            Argument:
                x {Tensor, shape: [batch, length, 2]} -- sequence position index masked
            Returns:
                x {Tensor, shape: [batch, length, d_features]} -- positional encoding
        '''
        facility_index = x[:, :, 1].long()
        facility_mask = facility_index == 0
        facility = self.facility_enc(facility_index)
        time = x[:, :, 0].masked_fill(facility_mask, 0)
        time = self.time_enc(time)

        x = time + facility
        return x


import gensim
import torch

def get_embeddings(model_path, padding = False):
    '''
        Arguments:
            model_path {String} -- gensim model file path
            padding {Boolean} -- Whether to expand a 0 vector for padding
        Returns:
            embedding {Tensor, shape [n_vocabulary, embedding length]} -- word embedding vectors
            vector_length {Int} -- embedding length
    '''
    # Load word2vec model
    word2vec = gensim.models.Word2Vec.load(model_path)
    vector_length = word2vec.vector_size
    embedding = torch.FloatTensor(word2vec.wv.vectors)
    # Leave the first embedding to 0 for padding
    if padding:
        empty_embedding = torch.zeros_like(embedding[0]).unsqueeze(0)
        embedding = torch.cat([empty_embedding, embedding])

    return embedding, vector_length

def one_hot_embedding(index, num_classes):
    diag = torch.eye(num_classes).byte()
    index = index.view(-1).tolist()
    return diag[index]


import numpy as np
from torch.optim.adam import Adam


class AdamOptimizer(Adam):
    '''A simple wrapper class for learning rate scheduling'''

    def __init__(self, parameters, n_warmup_steps, lr_max, betas=(0.9, 0.999)):
        '''lr of the first epoch: 1/exp(warmup_steps, 3/2)'''
        super(AdamOptimizer, self).__init__(filter(lambda x: x.requires_grad, parameters), lr=lr_max, betas=betas)
        self.n_warmup_steps = n_warmup_steps
        self.n_current_steps = 0
        # self.init_lr = np.power(d_model, -0.5)
        self.init_lr = lr_max * np.power(n_warmup_steps, 0.5)

    def step(self):
        "Step with the inner optimizer"
        self._update_learning_rate()
        super().step()

    def zero_grad(self):
        "Zero out the gradients by the inner optimizer"
        super().zero_grad()

    def _get_lr_scale(self):
        '''
            if step > warmup_steps, return the former one
            if step < warmup_steps, return the later one
            lr: up and down
        '''
        return np.min([
            np.power(self.n_current_steps, -0.5),
            np.power(self.n_warmup_steps, -1.5) * self.n_current_steps])

    def _update_learning_rate(self):
        ''' Learning rate scheduling per step '''

        self.n_current_steps += 1
        lr = self.init_lr * self._get_lr_scale()

        for param_group in self.param_groups:
            param_group['lr'] = lr

    def state_dict(self):
        return super().state_dict()



import numpy as np


def feature_shuffle_index(d_features, depth):
    '''
        Arguments:
            d_features {Int} -- the dimension of the features
            depth {Int} -- the depth of the expected output

        Returns:
            indexes {list} -- replicated and shuffled indexes, length = d_features*depth

    '''
    index = np.arange(0, d_features).tolist()

    if d_features % 2 == 1:
        dim_odd = 1
    else:
        dim_odd = 0
    return odd_even_shuffle(index, depth, dim_odd)


def odd_even_shuffle(index, depth, dim_odd=1):
    indexes = [] + index

    def odd_step(index):
        odds = index[0:-1:2]
        evens = index[1:-1:2]
        last = [index[-1]]
        return evens + odds + last

    def even_step(index):
        first = [index[0]]
        evens = index[1::2]
        odds = index[2::2]
        if dim_odd:
            return first + odds + evens
        else:
            return first + evens + odds

    while (depth > 1):
        if depth % 2 == 1:
            index = odd_step(index)
        else:
            index = even_step(index)

        indexes += index

        depth -= 1

    return indexes




import matplotlib.pyplot as plt
from sklearn.metrics import precision_recall_curve, auc
from sklearn.metrics import roc_auc_score, roc_curve

plt.rcParams['savefig.dpi'] = 300  # pixel
plt.rcParams['figure.dpi'] = 300  # resolution
plt.rcParams["figure.figsize"] = [5, 4] # figure size

def precision_recall(y_pred, y_test):
    precisions, recalls, thresholds = precision_recall_curve(y_true=y_test, probas_pred=y_pred)
    area = auc(recalls, precisions)
    return area, precisions, recalls, thresholds

def plot_pr_curve(recalls, precisions, auc, x_axis = 1):
    plt.rcParams['savefig.dpi'] = 300  # pixel
    plt.rcParams['figure.dpi'] = 300  # resolution
    plt.rcParams["figure.figsize"] = [5, 4]  # figure size

    plt.plot(recalls, precisions, color="darkorange", label='Precision-Recall curve (area = %0.3f)' % auc)
    plt.plot([1, 0], [0, 1], color='navy', lw=2, linestyle='--')
    plt.xlim([0.0, x_axis])
    plt.ylim([0.0, 1.05])
    plt.xlabel('recall')
    plt.ylabel('precision')
    plt.title('Precision-Recall curve')
    plt.legend(loc="lower right")
    plt.show()

def auc_roc(y_pred, y_test):
    auc = roc_auc_score(y_true=y_test, y_score=y_pred)
    fprs, tprs, thresholds = roc_curve(y_true=y_test, y_score=y_pred)
    return auc, fprs, tprs, thresholds

def plot_roc_curve(fprs, tprs, auc, x_axis = 1):

    plt.plot(fprs, tprs, color="darkorange", label='ROC curve (area = %0.3f)' % auc)
    plt.plot([0, 1], [0, 1], color='navy', lw=2, linestyle='--')
    plt.xlim([0.0, x_axis])
    plt.ylim([0.0, 1.05])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('Receiver operating characteristic')
    plt.legend(loc="lower right")
    plt.show()


import torch
import torch.utils.data as data

class BalanceSampler(data.sampler.Sampler):
    def __init__(self, y_train, indices, num_samples=None, replacement=True):
        labels = y_train[indices]
        labels = torch.tensor(labels)
        if labels.dim() != 1:
            labels = labels.flatten()
        class_count = labels.bincount()

        class_weight = 1.0 / class_count.double()
        self.weights = class_weight[labels]
        self.weights = torch.DoubleTensor(self.weights)
        self.indices = indices
        if num_samples == None:
            self.num_samples = len(indices)
        else: self.num_samples =num_samples
        self.replacement = replacement

    def __iter__(self):
        return (self.indices[i] for i in torch.multinomial(self.weights, self.num_samples, self.replacement))

    def __len__(self):
        return self.num_samples



from torch.optim.adam import Adam
from Training.losses import *
from Training.evaluation import accuracy, precision_recall, Evaluator
from Training.control import TrainingControl, EarlyStopping
from Modules.linear import LinearClassifier
from framework.model import Model
from tqdm import tqdm
from utils.plot_curves import precision_recall as pr
from utils.plot_curves import plot_pr_curve
from Modules.attention import *
from Layers.expansions import *


class ReactionModel(Model):
    def __init__(
            self, save_path, log_path, n_depth, d_features, d_meta, d_classifier, d_output, threshold,
            stack='ReactionAttention', expansion_layer='LinearExpansion', mode='1d',optimizer=None, **kwargs):
        '''*args: n_layers, n_head, dropout, use_bottleneck, d_bottleneck'''

        super().__init__(save_path, log_path)
        self.d_output = d_output
        self.threshold = threshold

        # ----------------------------- Model ------------------------------ #
        stack_dict = {
            'ReactionAttention': ReactionAttentionStack,
            'SelfAttention': SelfAttentionStack,
            'Alternate': AlternateStack,
            'Parallel': ParallelStack,
            'ShuffleSelfAttention': ShuffleSelfAttentionStack
        }
        expansion_dict = {
            'LinearExpansion': LinearExpansion,
            'ReduceParamLinearExpansion': ReduceParamLinearExpansion,
            'ConvExpansion': ConvExpansion,
            'LinearConvExpansion': LinearConvExpansion,
            'ShuffleConvExpansion': ShuffleConvExpansion,
            'ChannelWiseConvExpansion': ChannelWiseConvExpansion,
        }
        # *args:  n_head, n_depth,d_bottleneck=256, dropout=0.1, use_bottleneck=True
        self.model = stack_dict[stack](expansion_dict[expansion_layer],n_depth=n_depth, d_features=d_features, d_meta=d_meta, **kwargs)

        # --------------------------- Classifier --------------------------- #
        if mode == '1d':
            self.classifier = LinearClassifier(d_features, d_classifier, d_output)
        elif mode =='2d':
            self.classifier = LinearClassifier(n_depth * d_features, d_classifier, d_output)
        else:
            self.classifier = None


        self.CUDA_AVAILABLE = self.check_cuda()
        if self.CUDA_AVAILABLE:
            self.model.cuda()
            self.classifier.cuda()
        else:
            print('CUDA not found or not enabled, use CPU instead')

        # ---------------------------- Optimizer --------------------------- #
        self.parameters = list(self.model.parameters()) + list(self.classifier.parameters())
        if optimizer == None:
            self.optimizer = Adam(self.parameters, lr=0.001, betas=(0.9, 0.999))

        # ------------------------ training control ------------------------ #
        self.controller = TrainingControl(max_step=100000, evaluate_every_nstep=100, print_every_nstep=10)
        self.early_stopping = EarlyStopping(patience=50)


        # ---------------------------- END INIT ---------------------------- #

    def train_epoch(self, train_dataloader, eval_dataloader, device, smothing):
        ''' Epoch operation in training phase'''
        if device == 'cuda':
            assert self.CUDA_AVAILABLE
        # Set model and classifier training mode
        self.model.train()
        self.classifier.train()

        total_loss = 0
        batch_counter = 0

        # update param per batch
        for batch in tqdm(
                train_dataloader, mininterval=1,
                desc='  - (Training)   ', leave=False):  # training_data should be a iterable


            # get data from dataloader
            feature_1, feature_2, y = map(lambda x: x.to(device), batch)

            batch_size = len(feature_1)

            # forward
            self.optimizer.zero_grad()
            logits, attn = self.model(feature_1, feature_2)
            logits = logits.view(batch_size, -1)
            logits = self.classifier(logits)

            # Judge if it's a regression problem
            if self.d_output == 1:
                pred = logits.sigmoid()
                loss = mse_loss(pred, y)

            else:
                '''Do we need to apply softmax before calculating the loss?'''
                pred = logits
                loss = cross_entropy_loss(pred, y, smoothing=smothing)


            # calculate gradients
            loss.backward()

            # update parameters
            self.optimizer.step()

            # get metrics for logging
            acc = accuracy(pred, y, threshold=self.threshold)
            precision, recall, precision_avg, recall_avg = precision_recall(pred, y, self.d_output,
                                                                            threshold=self.threshold)
            total_loss += loss.item()
            batch_counter += 1

            # training control
            state_dict = self.controller(batch_counter)

            if state_dict['step_to_print']:
                self.train_logger.info(
                    '[TRAINING]   - batch: %5d, loss: %3.4f, acc: %1.4f, pre: %1.4f, rec: %1.4f' % (
                        batch_counter, loss, acc, precision[1], recall[1]))

            if state_dict['step_to_evaluate']:
                stop = self.val_epoch(eval_dataloader, device, state_dict['step'])
                state_dict['step_to_stop'] = stop
                if stop:
                    break

            if self.controller.current_step == self.controller.max_step:
                state_dict['step_to_stop'] = True
                break

        return state_dict

    def val_epoch(self,dataloader, device, step=0, plot=False):
        ''' Epoch operation in evaluation phase '''
        if device == 'cuda':
            assert self.CUDA_AVAILABLE

        # Set model and classifier training mode
        self.model.eval()
        self.classifier.eval()

        # use evaluator to calculate the average performance
        evaluator = Evaluator()

        pred_list = []
        real_list = []

        with torch.no_grad():

            for batch in tqdm(
                    dataloader, mininterval=5,
                    desc='  - (Evaluation)   ', leave=False):  # training_data should be a iterable


                # get data from dataloader
                feature_1, feature_2, y = map(lambda x: x.to(device), batch)
                batch_size = len(feature_1)

                # get logits
                logits, attn = self.model(feature_1, feature_2)
                logits = logits.view(batch_size, -1)
                logits = self.classifier(logits)

                if self.d_output == 1:
                    pred = logits.sigmoid()
                    loss = mse_loss(pred, y)

                else:
                    pred = logits
                    loss = cross_entropy_loss(pred, y, smoothing=False)

                acc = accuracy(pred, y, threshold=self.threshold)
                precision, recall, _, _ = precision_recall(pred, y, self.d_output, threshold=self.threshold)

                # feed the metrics in the evaluator
                evaluator(loss.item(), acc.item(), precision[1].item(), recall[1].item())

                '''append the results to the predict / real list for drawing ROC or PR curve.'''
                if plot:
                    pred_list += pred.tolist()
                    real_list += y.tolist()

            if plot:
                area, precisions, recalls, thresholds = pr(pred_list, real_list)
                plot_pr_curve(recalls, precisions, auc=area)

            # get evaluation results from the evaluator
            loss_avg, acc_avg, pre_avg, rec_avg = evaluator.avg_results()

            self.train_logger.info(
                '[EVALUATION] - eval_loss: %3.4f, acc: %1.4f, pre: %1.4f, rec: %1.4f' % (
                    loss_avg, acc_avg, pre_avg, rec_avg))

            state_dict = self.early_stopping(loss_avg)

            if state_dict['save']:
                checkpoint = {
                    'model_state_dict': self.model.state_dict(),
                    'classifier_state_dict': self.classifier.state_dict(),
                    'optimizer_state_dict': self.optimizer.state_dict(),
                    'global_step': step}
                self.save_model(checkpoint, self.save_path + '-loss-%.5f'%loss_avg)

            return state_dict['break']

    def train(self, epoch, train_dataloader, eval_dataloader, device, smoothing, save_mode):
        # set logger
        self.set_logger()

        assert save_mode in ['all', 'best']
        # train for n epoch
        for epoch_i in range(epoch):
            print('[ Epoch', epoch_i, ']')
            # set current epoch
            self.controller.set_epoch(epoch_i + 1)
            # train for on epoch
            state_dict = self.train_epoch(train_dataloader, eval_dataloader, device, smoothing)

            self.val_epoch(eval_dataloader, device, plot=True)

            if state_dict['step_to_stop']:
                break

        print('[INFO]: Finish Training, ends with %d epoch(s) and %d batches, in total %d training steps.' % (
            state_dict['epoch'] - 1, state_dict['batch'], state_dict['step']))

    def get_predictions(self, data_loader, device, max_batches=None, activation=None):

        pred_list = []
        real_list = []

        self.model.eval()
        self.classifier.eval()

        batch_counter = 0

        with torch.no_grad():
            for batch in tqdm(
                    data_loader,
                    desc='  - (Testing)   ', leave=False):

                # get data from dataloader
                feature_1, feature_2, y = map(lambda x: x.to(device), batch)

                # get logits
                logits, attn = self.model(feature_1, feature_2)
                logits = logits.view(logits.shape[0], -1)
                logits = self.classifier(logits)

                # Whether to apply activation function
                if activation != None:
                    pred = activation(logits)
                else:
                    pred = logits
                pred_list += pred.tolist()
                real_list += y.tolist()

                if max_batches != None:
                    batch_counter += 1
                    if batch_counter >= max_batches:
                        break

        return pred_list, real_list





import numpy as np

def padding_data(data, max_length, padding_value):
    '''
        padding a list of lists to max_length, and return a 2D list
        Arguments:
            data {Pandas Series} -- sequence to be padded, each element is a list of word index
            max_length {Int} -- the max length of the sequences
            padding_value {Int} -- the value to fill in
        Returns:
            data {2-D Numpy ndarray} -- a 2-D array with shape [number of sequences, max_length]
    '''

    def padding(example):
        '''padding the lists to the same length'''
        seq_length = len(example)
        return np.pad(example, (0, max_length - seq_length), constant_values=padding_value, mode='constant')
    data = data.map(padding)
    data = np.concatenate(data.values)
    data = data.reshape(-1, max_length) # Squeeze the lists to one list, and reshape

    return data



import torch
import torch.nn as nn
from torch.autograd import Variable


class LabelSmoothing(nn.Module):
    "Implement label smoothing."

    def __init__(self, size, padding_idx, smoothing=0.0):
        super(LabelSmoothing, self).__init__()
        self.criterion = nn.KLDivLoss(size_average=False)
        self.padding_idx = padding_idx
        self.confidence = 1.0 - smoothing
        self.smoothing = smoothing
        self.size = size
        self.true_dist = None

    def forward(self, x, target):
        assert x.size(1) == self.size
        true_dist = x.data.clone()
        true_dist.fill_(self.smoothing / (self.size - 2))
        true_dist.scatter_(1, target.data.unsqueeze(1), self.confidence)
        true_dist[:, self.padding_idx] = 0
        mask = torch.nonzero(target.data == self.padding_idx)
        if mask.dim() > 0:
            true_dist.index_fill_(0, mask.squeeze(), 0.0)
        self.true_dist = true_dist
        return self.criterion(x, Variable(true_dist, requires_grad=False))


import torch.nn as nn
from Layers.resnet import ResNetLayer
import numpy as np

class ResNet(nn.Module):
    ''' A encoder models with self attention mechanism. '''

    def __init__(
            self, position_encoding_layer, n_layers, n_head, d_features, max_seq_length, d_meta, d_k=None, d_v=None, dropout=0.1, use_bottleneck=True, d_bottleneck=256):
        super().__init__()

        if d_k == None or d_v == None:
            if d_k == d_v:
                d_reduce_param = np.floor(d_features / n_head).astype(int)
                d_k, d_v = d_reduce_param, d_reduce_param
            elif d_k == None:
                d_k = d_v
            else:
                d_v = d_k

        self.position_enc = position_encoding_layer(d_features=d_features, max_length=max_seq_length, d_meta=d_meta)

        self.layer_stack = nn.ModuleList([
            ResNet(d_features, n_head, d_k, d_v, dropout, use_bottleneck=use_bottleneck, d_bottleneck=d_bottleneck)
            for _ in range(n_layers)])

    def forward(self, feature_sequence, position, non_pad_mask=None, slf_attn_mask=None):
        '''
            Arguments:
                input_feature_sequence {Tensor, shape [batch, max_sequence_length, d_features]} -- input feature sequence
                position {Tensor, shape [batch, max_sequence_length (, d_meta)]} -- input feature position sequence
                non_pad_mask {Tensor, shape [batch, length, 1]} -- index of which position in a sequence is a padding
                slf_attn_mask {Tensor, shape [batch, length, length]} -- self attention mask

            Returns:
                enc_output {Tensor, shape [batch, max_sequence_length, d_features]} -- encoder output (representation)
                encoder_self_attn_list {List, length: n_layers} -- encoder self attention list,
                each element is a Tensor with shape [n_head * batch, max_sequence_length, max_sequence_length]
        '''

        encoder_self_attn_list = []
        enc_output = feature_sequence

        # Add position information at the beginning
        pos_enc = self.position_enc(position)
        enc_output = feature_sequence + pos_enc

        for enc_layer in self.layer_stack:
            enc_output, encoder_self_attn = enc_layer(
                enc_output,
                non_pad_mask=non_pad_mask,
                slf_attn_mask=slf_attn_mask)

            encoder_self_attn_list += [encoder_self_attn]

        return enc_output, encoder_self_attn_list


import torch.nn as nn

class LinearBottleneckLayer(nn.Module):
    ''' Bottleneck Layer '''

    def __init__(self, d_features, d_hid, d_out=None, dropout=0.1):
        super().__init__()
        if d_out == None:
            d_out = d_features

        self.encode = nn.Linear(d_features, d_hid)
        self.decode = nn.Linear(d_hid, d_out)
        nn.init.xavier_normal_(self.encode.weight)
        nn.init.xavier_normal_(self.decode.weight)
        self.layer_norm = nn.LayerNorm(d_features)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        '''
            Arguments:
                x {Tensor, shape [batch_size, d_features]}
            Returns:
                x {Tensor, shape [batch_size, d_features]}
        '''
        residual = x
        encode = nn.functional.relu(self.encode(x))
        decode = self.decode(encode)
        output = self.dropout(decode)
        output = self.layer_norm(output + residual)
        output = output + residual
        return output

class ShuffleBottleneckLayer(nn.Module):
    ''' Bottleneck Layer '''

    def __init__(self, n_depth, d_features, mode, d_hid=None, dropout=0.1):
        super().__init__()
        self.n_depth = n_depth
        self.d_features = d_features
        self.mode = mode
        if d_hid == None:
            d_hid = d_features

        if mode == '1d':
            self.bottle_neck_1 = nn.Linear(d_features, d_hid)
            self.bottle_neck_2 = nn.Linear(d_hid, d_features)

        elif mode == '2d':
            # self.bottle_neck_1 = nn.Conv1d(n_depth, d_hid, kernel_size=1, bias=False)
            # self.bottle_neck_2 = nn.Conv1d(d_hid, n_depth, kernel_size=1, bias=False)
            self.bottle_neck_1 = nn.Conv1d(d_features, d_hid, kernel_size=1)
            self.bottle_neck_2 = nn.Conv1d(d_hid, d_features, kernel_size=1)
        else:
            pass

        nn.init.xavier_normal_(self.bottle_neck_1.weight)
        nn.init.xavier_normal_(self.bottle_neck_2.weight)
        self.layer_norm = nn.LayerNorm([d_features])

        self.activation = nn.functional.relu

        self.dropout = nn.Dropout(dropout)

    def forward(self, features):
        '''
            Arguments:
                features {Tensor, shape [batch, d_features] or [batch, n_depth, d_features]} -- features

            Returns:
                x {Tensor, shape [batch_size, d_features]}
        '''
        residual = features

        if self.mode == '1d':
            output = self.bottle_neck_1(features)
            output = self.activation(output)
            output = self.bottle_neck_2(output)


        elif self.mode == '2d':
            output = features.transpose(1, 2)
            output = self.bottle_neck_1(output)
            output = self.activation(output)
            output = self.bottle_neck_2(output)
            output = output.transpose(2, 1)
        else:
            residual = 0
            output = features

        output = self.dropout(output)
        output = self.layer_norm(output + residual)

        return output

class ShuffleBottleneckLayerV2(nn.Module):
    ''' Bottleneck Layer '''

    def __init__(self, n_depth, d_features, mode, d_hid=None, dropout=0.1, c_total=3):
        super().__init__()
        self.n_depth = n_depth
        self.d_features = d_features
        self.mode = mode
        if d_hid == None:
            d_hid = d_features

        if mode == 'residual':
            self.bottle_neck_1 = nn.Conv1d(d_features, d_hid, kernel_size=1)
            self.bottle_neck_2 = nn.Conv1d(d_hid, d_features, kernel_size=1)

        elif mode == 'dense':
            self.bottle_neck_1 = nn.Conv1d(c_total, d_hid, kernel_size=1)
            self.bottle_neck_2 = nn.Conv1d(d_hid, c_total, kernel_size=1)
        else:
            pass

        nn.init.xavier_normal_(self.bottle_neck_1.weight)
        nn.init.xavier_normal_(self.bottle_neck_2.weight)
        self.layer_norm = nn.LayerNorm([d_features])

        self.activation = nn.functional.relu

        self.dropout = nn.Dropout(dropout)

    def forward(self, features):
        '''
            Arguments:
                features {Tensor, shape [batch, ?, d_features]} -- features

            Returns:
                x {Tensor, shape [batch_size, d_features]}
        '''
        residual = features

        if self.mode == 'residual':
            # features {Tensor, shape [batch, 3, d_features]}
            output = features.transpose(1, 2)
            output = self.bottle_neck_1(output)
            output = self.activation(output)
            output = self.bottle_neck_2(output)
            output = output.transpose(2, 1)

        elif self.mode == 'dense':
            # features {Tensor, shape [batch, c_total, d_features]}
            output = self.bottle_neck_1(features)  # [batch, d_hid, d_features]
            output = self.activation(output)
            output = self.bottle_neck_2(output)  # [batch, c_total, d_features]
            output = output.transpose(2, 1)
        else:
            residual = 0
            output = features

        output = self.dropout(output)
        output = self.layer_norm(output + residual)

        return output


import torch
import pandas as pd
from torch.utils.data import Dataset, DataLoader
from utils.padding import padding_data
from sklearn.model_selection import train_test_split


def process_indexes(index_series, max_length=None):
    '''
        Argument:
            index_series {Pandas Series} -- A Series of indexes, each element is a string contains the indexes
        Returns:
            idx_seq {Numpy ndarray, shape [n_rows_of_data, max_length]} -- a 2-D array of word indexes, padding with 0
            pos_seq {Numpy ndarray, shape [n_rows_of_data, max_length]} -- a 2-D array of position indexes, padding with 0
            max_length {Int} -- the max length of all lists in the Series
    '''
    # string to list
    # index_series = index_series.reset_index()

    indexes = index_series.map(lambda x: eval(x))
    # get list length
    length = indexes.map(lambda x: len(x))
    # generate position
    position = length.map(lambda x: list(range(0, x)))
    # get the max length
    max_length_ = length.max()
    # indexes {Pandas Series} -- each element is a list of indexes
    # length {Pandas Seires} -- each element is the length of the list of indexes
    # position {Pandas Series} -- each element is a list of positions
    # max_length {Int} -- the max length of all lists in the Series

    if max_length is None:
        max_length = max_length_
    else:
        if max_length < max_length_:
            raise Exception(
                'Exists index sequence with a length larger than max_length, please check if max_length applies for all of your datasets.')

    # Pad the sequences with -1 to the same length (max_length), for the use of indexing, shift the sequence with 1
    idx_seq = padding_data(indexes, max_length, padding_value=-1) + 1
    pos_seq = padding_data(position, max_length, padding_value=-1) + 1

    return idx_seq, pos_seq, max_length


class TransformerClsDataset(Dataset):
    def __init__(self, train_file_path, test_file_path, eval_size=0.1, max_length=None):
        super(TransformerClsDataset, self).__init__()
        # Load training data
        train_set = pd.read_csv(train_file_path)

        # Split eval set

        train_set, eval_set = train_test_split(train_set, test_size=eval_size, random_state=12)

        # generate dataset
        self.idx_seq_train, self.pos_seq_train, max_length = process_indexes(train_set['indexes'], max_length)

        self.target_train = train_set['is_vulnerable'].values.reshape(-1, 1)

        idx_seq_eval, pos_seq_eval, max_length = process_indexes(eval_set['indexes'], max_length)

        self.eval_set = (torch.from_numpy(idx_seq_eval), torch.from_numpy(pos_seq_eval),
                         torch.from_numpy(eval_set['is_vulnerable'].values.reshape(-1, 1)))

        # load test data

        test_set = pd.read_csv(test_file_path)

        # generate dataset
        idx_seq_test, pos_seq_test, _ = process_indexes(test_set['indexes'], max_length)

        self.test_set = (torch.from_numpy(idx_seq_test), torch.from_numpy(pos_seq_test),
                         torch.from_numpy(test_set['is_vulnerable'].values.reshape(-1, 1)))

        self.max_length = max_length

    def __len__(self):
        return len(self.target_train)

    def __getitem__(self, index):
        if torch.is_tensor(index):
            index = index.tolist()
        sample = (torch.from_numpy(self.idx_seq_train[index, :]),
                  torch.from_numpy(self.pos_seq_train[index, :]),
                  torch.from_numpy(self.target_train[index, :]))

        return sample





from numpy.lib.stride_tricks import as_strided as strided
import torch.as_strided

def get_sliding_windows(dataFrame, windowSize, return_flatten=False):
    '''
        Arguments:
            dataFrame {Pandas DataFrame} -- dataframe to be strided
            windowSize {Int} -- sliding window size
            return_flatten {bool} -- whether to flatten rows

        Returns:
            output {Numpy ndarray} -- strided data
    '''
    stride_row, stride_col = dataFrame.values.strides
    rows, columns = dataFrame.shape
    output = strided(dataFrame, shape=(rows - windowSize + 1, windowSize, columns),
                     strides=(stride_row, stride_row, stride_col))
    if return_flatten == 1:
        return output.reshape(dataFrame.shape[0] - windowSize + 1, -1)
    else:
        return output


import torch
import torch.nn as nn

class LinearClassifier(nn.Module):
    def __init__(self, d_in, d_hid, d_out):
        super(LinearClassifier, self).__init__()

        self.fc1 = nn.Linear(d_in, d_hid)
        self.activation = nn.functional.leaky_relu
        self.fc2 = nn.Linear(d_hid, d_out)

        nn.init.xavier_normal_(self.fc1.weight)
        nn.init.xavier_normal_(self.fc2.weight)

    def forward(self, x):
        x = self.fc1(x)
        x = self.activation(x)
        x = self.fc2(x)
        return x


import torch
import torch.nn as nn
import numpy as np

class ResNetLayer(nn.Module):
    # def __init__(self, d_model, d_inner, n_head, d_k, d_v, dropout=0.1):
    def __init__(self, in_channel, out_channel, stride,
                 use_bottleneck=False, d_bottleneck=128):
        super().__init__()
        self.bn1 = nn.BatchNorm2d(in_channel)
        self.act1 = nn.ReLU()
        self.conv1 = nn.Conv2d(in_channel, out_channel, kernel_size=3, stride=stride, padding=)
        self.bn2 = nn.BatchNorm2d(out_channel)

    def forward(self, enc_input, non_pad_mask=None, slf_attn_mask=None):
        '''
            Arguments:
                enc_input {Tensor, shape [batch, length, d_features]} -- input
                non_pad_mask {Tensor, shape [batch, length, 1]} -- index of which position in a sequence is a padding
                slf_attn_mask {Tensor, shape [batch, length, length]} -- self attention mask

            Returns:
                enc_output {Tensor, shape [batch, length, d_features]} -- output
                encoder_self_attn {n_head * batch, length, length} -- n_head self attention matrices
        '''

        enc_output, encoder_self_attn = self.self_attn(
            enc_input, enc_input, enc_input, mask=slf_attn_mask)
        enc_output *= non_pad_mask

        enc_output = self.bottleneck(enc_output)
        enc_output *= non_pad_mask

        return enc_output, encoder_self_attn


import torch
from torch.utils.data import Dataset, DataLoader
from torch.utils.data.sampler import SubsetRandomSampler
from Dataloader.samplers import BalanceSampler
import numpy as np


class ReactionDataset(Dataset):
    def __init__(self, f1_path, f2_path, y_path):
        super(ReactionDataset, self).__init__()

        # load data
        self.feature1 = np.load(f1_path)
        self.feature2 = np.load(f2_path)
        self.y = np.load(y_path)

    def __len__(self):
        return len(self.y)

    def __getitem__(self, index):
        if torch.is_tensor(index):
            index = index.tolist()
        sample = (torch.from_numpy(self.feature1[index, :]).float(),
                  torch.from_numpy(self.feature2[index, :]).float(),
                  torch.from_numpy(self.y[index, :]).float())

        return sample

    def get_feature_dim(self):
        return self.feature1.shape[-1], self.feature2.shape[-1]


class ReactionDataloader():
    def __init__(self, file_dir, batch_size, val_size, shuffle=True):
        # feature1 shape [?, n_features]
        # feature2 shape [?, m_features]
        # y shape [?, 1]

        # Load training data
        train_set = ReactionDataset(file_dir + 'X_train.npy', file_dir + 'dev_train.npy', file_dir + 'y_train.npy')
        self.dims = train_set.get_feature_dim()

        train_size = len(train_set)
        indices = list(range(train_size))
        split = int(np.floor(val_size * train_size))
        if shuffle:
            np.random.seed(42)
            np.random.shuffle(indices)
        train_indices, val_indices = indices[split:], indices[:split]

        train_sampler = SubsetRandomSampler(train_indices)
        valid_sampler = SubsetRandomSampler(val_indices)

        # train_sampler = BalanceSampler(train_set.y, train_indices)
        # valid_sampler = BalanceSampler(train_set.y, val_indices)


        self.train_loader = DataLoader(train_set, batch_size, sampler=train_sampler, num_workers=4)
        self.val_loader = DataLoader(train_set, batch_size, sampler=valid_sampler, num_workers=4)

        # Load test data
        test_set = ReactionDataset(file_dir + 'X_test.npy', file_dir + 'dev_test.npy', file_dir + 'y_test.npy')

        self.test_loader = DataLoader(test_set, batch_size, num_workers=4)

    def train_dataloader(self):
        return self.train_loader

    def val_dataloader(self):
        return self.val_loader

    def test_dataloader(self):
        return self.test_loader

    def get_feature_dim(self):
        return self.dims
